/*	$NetBSD: xxboot.c,v 1.4 2001/03/26 11:54:50 tsutsui Exp $ */

/*-
 * Copyright (c) 1982, 1986, 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * 	@(#)boot.c	8.1 (Berkeley) 6/10/93
 */


#ifndef ELFSIZE
#define	ELFSIZE	32
#endif

#include <sys/param.h>
#include <sys/reboot.h>
#include <machine/mon.h>

#include <stand.h>
#include <loadfile.h>
#include "libsa.h"

const char bootprog_rev[] = "05/11/2006";

/*
 * Note that extname is edited based on the running machine type
 * (sun3 vs sun3x).  EXTNAMEX is the position of the 'x'.
 */
char	extname[] = "vmlinux.sun3x";
#define EXTNAMEX (sizeof(extname)-2)

/*
 * If the PROM did not give us a specific kernel name to use,
 * and did not specify the -a flag (ask), then try the names
 * in the following list.
 */
char *kernelnames[] = {
	"vmlinux",
	"vmlinux.old",
	extname,
	NULL
};
char	line[180];

extern void *init_bootinfo(void *start, char *params, void *rddata, int rdsize, unsigned long machtype);

/* code to disable interrupts on the 3x -- this is needed to boot properly */
#define  OBIO_INTERREG           0x61001400
extern u_int sun3x_ptov(u_int pa);
void sun3x_disable_interrupts(void)
{
	u_int *intreg = (u_int *)sun3x_ptov(OBIO_INTERREG);

	*intreg = 0;
}

void
xxboot_main(const char *boot_type)
{
	struct open_file	f;
	char **npp;
	char *file;
	char *cmdline;
	void *entry;
	int fd, i;
	u_long marks[MARK_MAX];
	Elf_Ehdr *ehdr;
	Elf_Shdr *shdr;
	unsigned long end = 0;
	unsigned long offset = 0;
	unsigned long *bi_size;

	memset(marks, 0, sizeof(marks));
	printf(">> Linux/sun3 %s [%s]\n", boot_type, bootprog_rev);
	prom_get_boot_info();

	/*
	 * Hold the raw device open so it will not be
	 * closed and reopened on every attempt to
	 * load files that did not exist.
	 */
	f.f_flags = F_RAW;
	if (devopen(&f, 0, &file)) {
		printf("%s: devopen failed\n", boot_type);
		return;
	}

	/*
	 * Edit the "extended" kernel name based on
	 * the type of machine we are running on.
	 */
	if (_is3x == 0)
		extname[EXTNAMEX] = 0;

	/*
	 * If the PROM gave us a file name,
	 * it means the user asked for that
	 * kernel name explicitly.
	 */
	file = prom_bootfile;
	cmdline = prom_bootargs;
	if (file && *file) {
		fd = loadfile(file, marks, LOAD_KERNEL);
		if (fd == -1) {
			goto err;
		} else {
			goto gotit;
		}
	}

	/*
	 * Try the default kernel names.
	 */
	for (npp = kernelnames; *npp; npp++) {
		file = *npp;
		printf("%s: trying %s\n", boot_type, file);
		fd = loadfile(file, marks, LOAD_KERNEL);
		if (fd != -1)
			goto gotit;
	}

	/*
	 * Ask what kernel name to load.
	 */
	for (;;) {

		file = kernelnames[0];
		cmdline = NULL;

		printf("filename? [%s]: ", file);
		gets(line);
		if (line[0]) {
			file = line;
			cmdline = (char *)strchr(line, ' ');
			if(cmdline != NULL) {
				*cmdline = 0;
				cmdline++;
			}
		}

		fd = loadfile(file, marks, LOAD_KERNEL);
		if (fd != -1)
			break;

	err:
		printf("%s: %s: loadfile() failed.\n", boot_type, file);
	}

gotit:
#ifdef DEBUG
	printf("mark_start 0x%x mark_sym 0x%x mark_end 0x%x\n", marks[MARK_START], marks[MARK_SYM], marks[MARK_END]);
#endif
	
	/* now, find the end of allocated ram -- this is where _end will be */
	end = 0;
	ehdr = (Elf_Ehdr *)marks[MARK_SYM];
	shdr = (Elf_Shdr *)(marks[MARK_SYM] + ehdr->e_shoff);

	printf("shnum %d\n", ehdr->e_shnum);
	for(i = 0; i < ehdr->e_shnum; i++) {
		unsigned long x;
		
		if(shdr[i].sh_flags & SHF_ALLOC) {
			x = shdr[i].sh_addr + shdr[i].sh_size;
			if(x > end) 
				end = x;
		}
	}

	end = LOADADDR(end);
	printf("end: 0x%x\n", end);

	init_bootinfo((void *)end, cmdline, NULL, 0, _is3x);

	entry = (void *)marks[MARK_ENTRY];
	printf("Starting program at 0x%x\n", entry);
	
	if(_is3x)
		sun3x_disable_interrupts();

	chain_to(entry);
}
